﻿using Microsoft.Win32.SafeHandles;
using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Globalization;
using System.IO;
using System.IO.Compression;
using System.Linq;
using System.Net;
using System.Runtime.ConstrainedExecution;
using System.Runtime.InteropServices;
using System.Security.Permissions;
using System.Text;

namespace WinInetDemo3
{
    [SecurityPermission(SecurityAction.LinkDemand, UnmanagedCode = true)]
    internal class MySafeInetHandle : SafeHandleZeroOrMinusOneIsInvalid
    {
        internal MySafeInetHandle()
            : base(true)
        {
        }

        [ReliabilityContract(Consistency.WillNotCorruptState, Cer.MayFail)]
        protected override bool ReleaseHandle()
        {
            bool flag = NativeMethods.InternetCloseHandle(this.handle);
            if (!flag)
            {
                Marshal.GetLastWin32Error();
            }
            return flag;
        }
    }

    internal class MyBytesReadEventArgs : EventArgs
    {
        private int _bytesRead;

        private int _totalBytes;

        internal int BytesRead
        {
            get
            {
                return this._bytesRead;
            }
        }

        internal int TotalBytes
        {
            get
            {
                return this._totalBytes;
            }
        }

        internal MyBytesReadEventArgs(int bytesRead, int totalBytes)
        {
            this._bytesRead = bytesRead;
            this._totalBytes = totalBytes;
        }
    }

    internal class MyWinInet : IDisposable
    {
        private class MyInternetReadStream : Stream
        {
            internal delegate void BytesReadCallback(int bytesRead, int totalBytes);

            private const int _localBufferSize = 32768;

            private readonly IntPtr _hInetFile;

            private byte[] _localBuffer;

            private readonly int _contentLength;

            private MyWinInet.MyInternetReadStream.BytesReadCallback _bytesReadCallback;

            public override bool CanRead
            {
                get
                {
                    return true;
                }
            }

            public override bool CanSeek
            {
                get
                {
                    return false;
                }
            }

            public override bool CanWrite
            {
                get
                {
                    return false;
                }
            }

            public override long Length
            {
                get
                {
                    throw new NotSupportedException();
                }
            }

            public override long Position
            {
                get
                {
                    throw new NotSupportedException();
                }
                set
                {
                    throw new NotSupportedException();
                }
            }

            internal MyInternetReadStream(IntPtr hInetFile, MyWinInet.MyInternetReadStream.BytesReadCallback bytesReadCallback)
            {
                this._hInetFile = hInetFile;
                this._localBuffer = new byte[BufferSize];
                this._bytesReadCallback = bytesReadCallback;
                this._contentLength = MyWinInet.GetContentLength(hInetFile);
            }

            public override void Flush()
            {
                throw new NotSupportedException();
            }

            public override int Read(byte[] buffer, int offset, int count)
            {
                if (buffer == null)
                {
                    throw new ArgumentNullException("buffer");
                }
                if (buffer.Length < offset + count)
                {
                    throw new ArgumentOutOfRangeException("count");
                }
                int dwNumberOfBytesToRead = Math.Min(BufferSize, count);
                int length = 0;
                if (!NativeMethods.InternetReadFile(this._hInetFile, this._localBuffer, dwNumberOfBytesToRead, out length))
                {
                    // todo
                }
                Array.Copy(this._localBuffer, 0, buffer, offset, length);
                this._bytesReadCallback(length, this._contentLength);
                return length;
            }

            public override long Seek(long offset, SeekOrigin origin)
            {
                throw new NotSupportedException();
            }

            public override void SetLength(long value)
            {
                throw new NotSupportedException();
            }

            public override void Write(byte[] buffer, int offset, int count)
            {
                throw new NotSupportedException();
            }

            protected override void Dispose(bool disposing)
            {
                if (disposing)
                {
                    this._bytesReadCallback = null;
                    this._localBuffer = null;
                }
                base.Dispose(disposing);
            }
        }

        private const int BufferSize = 32768;
        private EventHandler<MyBytesReadEventArgs> _bytesRead;
        private MySafeInetHandle _hInet;
        private bool _cancelDownload;
        private bool _disposed;

        internal event EventHandler<MyBytesReadEventArgs> BytesRead
        {
            add
            {
                this._bytesRead = (EventHandler<MyBytesReadEventArgs>)Delegate.Combine(this._bytesRead, value);
            }
            remove
            {
                this._bytesRead = (EventHandler<MyBytesReadEventArgs>)Delegate.Remove(this._bytesRead, value);
            }
        }

        internal bool CancelDownload
        {
            get
            {
                return this._cancelDownload;
            }
            set
            {
                this._cancelDownload = value;
            }
        }

        internal static string Referer
        {
            get
            {
                return "test referer";
            }
        }

        internal static string UserAgent
        {
            get
            {
                return "test user agent";
            }
        }

        internal MyWinInet()
        {
        }

        internal void InternetClose()
        {
            this._hInet.Dispose();
            this._hInet.SetHandleAsInvalid();
        }

        internal void InternetOpen()
        {
            this._hInet = NativeMethods.InternetOpen(MyWinInet.UserAgent, 0, null, null, 0);
            if (this._hInet.IsInvalid)
            {
                // todo
            }
        }

        internal bool DownloadFile(Uri uri, string fileName)
        {
            bool isDownloadSuccess = true;
            string contentDispositionFileName = null;
            if (this._hInet.IsInvalid)
            {
                throw new InvalidOperationException("InternetOpen has not been called yet");
            }
            string directoryName = Path.GetDirectoryName(fileName);
            if (!Directory.Exists(directoryName))
            {
                throw new ArgumentException("Directory '" + directoryName + "' does not exist");
            }
            IntPtr response = IntPtr.Zero;
            try
            {
                isDownloadSuccess = this.CheckResponseStatus(ref uri, ref response);
                if (isDownloadSuccess)
                {
                    byte[] buffer = new byte[BufferSize];
                    int contentLength = MyWinInet.GetContentLength(response);
                    contentDispositionFileName = MyWinInet.GetFileName(contentDispositionFileName, response, buffer);
                    using (FileStream fileStream = File.Open(fileName, FileMode.Create, FileAccess.Write, FileShare.None))
                    {
                        using (Stream internetStream = this.GetInternetStream(response))
                        {
                            int count;
                            while ((count = internetStream.Read(buffer, 0, BufferSize)) != 0 && !this._cancelDownload)
                            {
                                fileStream.Write(buffer, 0, count);
                            }
                        }
                    }
                }
            }
            finally
            {
                if (response != IntPtr.Zero)
                {
                    NativeMethods.InternetCloseHandle(response);
                    response = IntPtr.Zero;
                }
            }
            return isDownloadSuccess;
        }

        private Stream GetInternetStream(IntPtr hInetFile)
        {
            string contentEncoding = MyWinInet.GetContentEncoding(hInetFile);
            if (contentEncoding.IndexOf("gzip", StringComparison.OrdinalIgnoreCase) != -1)
            {
                return new GZipStream(this.ForGZipReadStream(hInetFile), CompressionMode.Decompress, false);
            }
            return new MyWinInet.MyInternetReadStream(hInetFile, new MyWinInet.MyInternetReadStream.BytesReadCallback(this.BytesReadCallback));
        }

        private Stream ForGZipReadStream(IntPtr hInetFile)
        {
            return new MyWinInet.MyInternetReadStream(hInetFile, new MyWinInet.MyInternetReadStream.BytesReadCallback(this.BytesReadCallback));
        }

        private bool CheckResponseStatus(ref Uri uri, ref IntPtr hInetFile)
        {
            string tempStr;
            byte[] content = new byte[BufferSize];
            string referer = "Referer: " + MyWinInet.Referer;
            referer += "\nAccept-Encoding: gzip";
            // INTERNET_FLAG_RELOAD
            // 0x80000000  -> 2147483648
            // Forces a download of the requested file, from the origin server, not from the cache. 
            hInetFile = NativeMethods.InternetOpenUrl(this._hInet, uri.AbsoluteUri, referer, referer.Length, 2149580800, IntPtr.Zero);
            if (hInetFile == IntPtr.Zero)
            {
                // todo
            }
            int count = BufferSize;
            int temp = 0;
            if (!NativeMethods.HttpQueryInfo(hInetFile, 19, content, out count, out temp))
            {
                // todo
            }
            tempStr = Encoding.Unicode.GetString(content, 0, count);
            if (!string.Equals(tempStr, "200", StringComparison.Ordinal))
            {
                return false;
            }
            return true;
        }

        private static string GetContentEncoding(IntPtr hInetFile)
        {
            int length1 = 0;
            int length2 = 100;
            byte[] content = new byte[length2];
            string text = string.Empty;
            try
            {
                if (NativeMethods.HttpQueryInfo(hInetFile, 29, content, out length2, out length1))
                {
                    text = Encoding.Unicode.GetString(content, 0, length2);
                }
            }
            catch
            {
            }
            return text;
        }

        private static string GetFileName(string contentDispositionFileName, IntPtr hInetFile, byte[] buffer)
        {
            try
            {
                int count = BufferSize;
                int length1 = 0;
                if (!NativeMethods.HttpQueryInfo(hInetFile, 47u, buffer, out count, out length1))
                {
                    Marshal.GetLastWin32Error();
                }
                else
                {
                    string tempStr = Encoding.Unicode.GetString(buffer, 0, count);
                    int length2 = tempStr.IndexOf("filename=", StringComparison.OrdinalIgnoreCase);
                    if (length2 > -1 && tempStr.Length > length2 + 9)
                    {
                        length2 += 9;
                        int length3 = tempStr.IndexOf(';', length2);
                        int length;
                        if (length3 > -1)
                        {
                            length = length3 - length2;
                        }
                        else
                        {
                            length = tempStr.Length - length2;
                        }
                        contentDispositionFileName = tempStr.Substring(length2, length).Trim(new char[]
                        {
                            '"'
                        });
                    }
                }
            }
            catch
            {
            }
            return contentDispositionFileName;
        }

        private static int GetContentLength(IntPtr hInetFile)
        {
            int result = 1;
            try
            {
                int num = 100;
                byte[] array = new byte[num];
                int num2 = 0;
                if (NativeMethods.HttpQueryInfo(hInetFile, 5u, array, out num, out num2))
                {
                    string tempStr = Encoding.Unicode.GetString(array, 0, num);
                    result = int.Parse(tempStr, CultureInfo.InvariantCulture);
                }
            }
            catch
            {
            }
            return result;
        }

        //private static bool HttpQueryInfoWithRetry(IntPtr hInetFile, uint dwInfoLevel, out byte[] lpvBuffer, out int dwBufferLength, out int dwIndex)
        //{
        //    dwBufferLength = 0;
        //    lpvBuffer = new byte[dwBufferLength];
        //    if (NativeMethods.HttpQueryInfo(hInetFile, dwInfoLevel, lpvBuffer, out dwBufferLength, out dwIndex))
        //    {
        //        return true;
        //    }
        //    lpvBuffer = new byte[dwBufferLength];
        //    if (NativeMethods.HttpQueryInfo(hInetFile, dwInfoLevel, lpvBuffer, out dwBufferLength, out dwIndex))
        //    {
        //        return true;
        //    }
        //    return false;
        //}

        private void BytesReadCallback(int bytesReadLength, int totalBytesLength)
        {
            EventHandler<MyBytesReadEventArgs> tempHandler = this._bytesRead;
            if (tempHandler != null)
            {
                tempHandler(this, new MyBytesReadEventArgs(bytesReadLength, totalBytesLength));
            }
        }

        public void Dispose()
        {
            this.Dispose(true);
            GC.SuppressFinalize(this);
        }

        protected virtual void Dispose(bool disposing)
        {
            if (!this._disposed)
            {
                if (this._hInet != null && !this._hInet.IsInvalid)
                {
                    this._hInet.Dispose();
                }
                this._disposed = true;
            }
        }

        ~MyWinInet()
        {
            this.Dispose(false);
        }
    }
}

